
@file:Suppress("unused", "UNUSED_PARAMETER")

package ru.dbotthepony.mc.otm

import net.minecraft.world.level.Level
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.event.TickEvent.WorldTickEvent
import net.minecraftforge.event.server.ServerAboutToStartEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import net.minecraftforge.eventbus.api.EventPriority
import net.minecraftforge.eventbus.api.SubscribeEvent
import java.util.*
import kotlin.collections.ArrayList

private val preServerTick = ArrayList<IConditionalTickable>()
private val postServerTick = ArrayList<IConditionalTickable>()
private val preServerTickOnce = ArrayList<ITickable>()
private val postServerTickOnce = ArrayList<ITickable>()

private val preWorldTick = WeakHashMap<Level, ArrayList<IConditionalTickable>>()
private val postWorldTick = WeakHashMap<Level, ArrayList<IConditionalTickable>>()
private val preWorldTickOnce = WeakHashMap<Level, ArrayList<ITickable>>()
private val postWorldTickOnce = WeakHashMap<Level, ArrayList<ITickable>>()

fun interface ITickable {
	fun tick()
}

interface IConditionalTickable : ITickable {
	/**
	 * Once this returns false, it should stay false.
	 *
	 * If it suddenly turns true after being false, result is undefined.
	 */
	val canTick: Boolean
}

@SubscribeEvent(priority = EventPriority.LOWEST)
fun onServerTick(event: ServerTickEvent) {
	if (event.phase === TickEvent.Phase.START) {
		for (i in preServerTick.size - 1 downTo 0) {
			val ticker = preServerTick[i]

			if (!ticker.canTick) {
				preServerTick.removeAt(i)
			} else {
				ticker.tick()
			}
		}

		for (i in preServerTickOnce.size - 1 downTo 0) {
			preServerTickOnce[i].tick()
			preServerTickOnce.removeAt(i)
		}
	} else {
		for (i in postServerTick.size - 1 downTo 0) {
			val ticker = postServerTick[i]

			if (!ticker.canTick) {
				postServerTick.removeAt(i)
			} else {
				ticker.tick()
			}
		}

		for (i in postServerTickOnce.size - 1 downTo 0) {
			postServerTickOnce[i].tick()
			postServerTickOnce.removeAt(i)
		}
	}
}

fun addPreServerTicker(ticker: IConditionalTickable) {
	preServerTick.add(ticker)
}

fun addPostServerTicker(ticker: IConditionalTickable) {
	postServerTick.add(ticker)
}

fun addPreServerTickerOnce(ticker: ITickable) {
	preServerTickOnce.add(ticker)
}

fun addPostServerTickerOnce(ticker: ITickable) {
	postServerTickOnce.add(ticker)
}

@SubscribeEvent(priority = EventPriority.LOWEST)
fun onWorldTick(event: WorldTickEvent) {
	if (event.phase === TickEvent.Phase.START) {
		val it = preWorldTick[event.world]

		if (it != null) {
			for (i in it.size - 1 downTo 0) {
				val ticker = it[i]

				if (!ticker.canTick) {
					it.removeAt(i)
				} else {
					ticker.tick()
				}
			}
		}

		val it2 = preWorldTickOnce.remove(event.world)

		if (it2 != null) {
			for (i in it2.size - 1 downTo 0) {
				it2[i].tick()
				it2.removeAt(i)
			}
		}
	} else {
		val it = postWorldTick[event.world]

		if (it != null) {
			for (i in it.size - 1 downTo 0) {
				val ticker = it[i]

				if (!ticker.canTick) {
					it.removeAt(i)
				} else {
					ticker.tick()
				}
			}
		}

		val it2 = postWorldTickOnce.remove(event.world)

		if (it2 != null) {
			for (i in it2.size - 1 downTo 0) {
				it2[i].tick()
				it2.removeAt(i)
			}
		}
	}
}

fun addPreWorldTicker(level: Level, ticker: IConditionalTickable) {
	preWorldTick.computeIfAbsent(level) { ArrayList() }.add(ticker)
}

fun addPostWorldTicker(level: Level, ticker: IConditionalTickable) {
	postWorldTick.computeIfAbsent(level) { ArrayList() }.add(ticker)
}

fun addPreWorldTickerOnce(level: Level, ticker: ITickable) {
	preWorldTickOnce.computeIfAbsent(level) { ArrayList() }.add(ticker)
}

fun addPostWorldTickerOnce(level: Level, ticker: ITickable) {
	postWorldTickOnce.computeIfAbsent(level) { ArrayList() }.add(ticker)
}

private fun clear() {
	preServerTick.clear()
	postServerTick.clear()
	preWorldTick.clear()
	postWorldTick.clear()
	preServerTickOnce.clear()
	postServerTickOnce.clear()
	preWorldTickOnce.clear()
	postWorldTickOnce.clear()
}

@SubscribeEvent(priority = EventPriority.HIGHEST)
fun onServerStarting(event: ServerAboutToStartEvent?) {
	clear()
}

@SubscribeEvent(priority = EventPriority.HIGHEST)
fun onServerStopping(event: ServerStoppingEvent?) {
	clear()
}
